"""Command-line interface for korgalore."""

import os
import click
import tomllib
import logging
import click_log

from pathlib import Path
from typing import Dict, Any, List, Tuple, Optional
from korgalore.gmail_service import GmailService
from korgalore.lore_service import LoreService
from korgalore.lei_service import LeiService
from korgalore import __version__, ConfigurationError, StateError, GitError, RemoteError

logger = logging.getLogger('korgalore')
click_log.basic_config(logger)

def get_xdg_data_dir() -> Path:
    # Get XDG_DATA_HOME or default to ~/.local/share
    xdg_data_home = os.environ.get('XDG_DATA_HOME')
    if xdg_data_home:
        data_home = Path(xdg_data_home)
    else:
        data_home = Path.home() / '.local' / 'share'

    # Create korgalore subdirectory
    korgalore_data_dir = data_home / 'korgalore'

    # Create directory if it doesn't exist
    korgalore_data_dir.mkdir(parents=True, exist_ok=True)

    return korgalore_data_dir


def get_xdg_config_dir() -> Path:
    # Get XDG_CONFIG_HOME or default to ~/.config
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME')
    if xdg_config_home:
        config_home = Path(xdg_config_home)
    else:
        config_home = Path.home() / '.config'

    # Create korgalore subdirectory
    korgalore_config_dir = config_home / 'korgalore'

    # Create directory if it doesn't exist
    korgalore_config_dir.mkdir(parents=True, exist_ok=True)

    return korgalore_config_dir


def get_target(ctx: click.Context, identifier: str) -> Any:
    if identifier in ctx.obj['targets']:
        return ctx.obj['targets'][identifier]

    config = ctx.obj.get('config', {})
    targets = config.get('targets', {})
    if identifier not in targets:
        logger.critical('Target "%s" not found in configuration.', identifier)
        logger.critical('Known targets: %s', ', '.join(targets.keys()))
        raise click.Abort()

    details = targets[identifier]
    if details.get('type') != 'gmail':
        logger.critical('Target "%s" is not a Gmail target.', identifier)
        raise click.Abort()

    gs = get_gmail_service(identifier=identifier,
                            credentials_file=details.get('credentials', ''),
                            token_file=details.get('token', None))
    ctx.obj['targets'][identifier] = gs
    return gs


def get_gmail_service(identifier: str, credentials_file: str,
                      token_file: Optional[str]) -> GmailService:
    if not credentials_file:
        logger.critical('No credentials file specified for Gmail target: %s', identifier)
        raise click.Abort()
    if not token_file:
        cfgdir = get_xdg_config_dir()
        token_file = str(cfgdir / f'gmail-{identifier}-token.json')
    try:
        gmail_service = GmailService(identifier=identifier,
                                     credentials_file=credentials_file,
                                     token_file=token_file)
    except ConfigurationError as fe:
        logger.critical('Error: %s', str(fe))
        raise click.Abort()

    return gmail_service


def load_config(cfgfile: Path) -> Dict[str, Any]:
    config: Dict[str, Any] = dict()

    if not cfgfile.exists():
        logger.error('Config file not found: %s', str(cfgfile))
        click.Abort()

    try:
        logger.debug('Loading config from %s', str(cfgfile))

        with open(cfgfile, 'rb') as cf:
            config = tomllib.load(cf)

        logger.debug('Config loaded with %s targets and %s sources',
                     len(config.get('targets', {})), len(config.get('sources', {})))

        return config

    except Exception as e:
        logger.error('Error loading config: %s', str(e))
        raise click.Abort()


def process_commits(listname: str, commits: List[str], gitdir: Path,
                    ctx: click.Context, max_count: int = 0) -> Tuple[int, str]:
    if max_count > 0 and len(commits) > max_count:
        # Take the last NN messages and discard the rest
        logger.info('Limiting to %d messages as requested', max_count)
        commits = commits[-max_count:]

    ls = ctx.obj['lore']
    cfg = ctx.obj.get('config', {})

    details = cfg['sources'][listname]
    target = details.get('target', '')
    try:
        gs = get_target(ctx, target)
    except click.Abort:
        logger.critical('Failed to process list "%s".', listname)
        raise ConfigurationError()

    last_commit = ''

    if logger.isEnabledFor(logging.DEBUG):
        hidden = True
    elif logger.isEnabledFor(logging.INFO):
        hidden = False
    else:
        hidden = True

    count = 0
    with click.progressbar(commits,
                            label=f'Uploading {listname}',
                            show_pos=True,
                            hidden=hidden) as bar:
        for at_commit in bar:
            try:
                raw_message = ls.get_message_at_commit(gitdir, at_commit)
            except (StateError, GitError) as e:
                logger.debug('Skipping commit %s: %s', at_commit, str(e))
                # Assuming non-m commit
                continue
            try:
                gs.import_message(raw_message, labels=details.get('labels', []))
                count += 1
            except RemoteError as re:
                logger.critical('Failed to upload message at commit %s: %s', at_commit, str(re))
                return count, last_commit
            ls.update_korgalore_info(gitdir=gitdir, latest_commit=at_commit, message=raw_message)
            last_commit = at_commit
            if logger.isEnabledFor(logging.DEBUG):
                msg = ls.parse_message(raw_message)
                logger.debug(' -> %s', msg.get('Subject', '(no subject)'))

    return count, last_commit


def process_lei_list(ctx: click.Context, listname: str,
                     details: Dict[str, Any], max_mail: int) -> int:
    # Make sure lei knows about this list
    # Placeholder for future LEI feed processing logic
    lei = ctx.obj['lei']
    if lei is None:
        lei = LeiService()
        ctx.obj['lei'] = lei
    feed = details.get('feed', '')[4:]  # Strip 'lei:' prefix
    if feed not in lei.known_searches:
        logger.critical('LEI search "%s" not known. Please create it first.', listname)
        raise click.Abort()
    feedpath = Path(feed)
    latest_epochs = lei.get_latest_epoch_info(feedpath)
    latest_epoch = max(lei.find_epochs(feedpath))
    try:
        known_epochs = lei.load_known_epoch_info(feedpath)
    except StateError:
        lei.save_epoch_info(list_dir=feedpath, epochs=latest_epochs)
        lei.update_korgalore_info(gitdir=feedpath / 'git' / f'{latest_epoch}.git')
        logger.info('Initialized: %s.', listname)
        return 0
    logger.debug('Running lei-up on list: %s', listname)
    lei.up_search(lei_name=feed)
    latest_epochs = lei.get_latest_epoch_info(feedpath)
    if known_epochs == latest_epochs:
        logger.debug('No updates for LEI list: %s', listname)
        return 0
    # XXX: this doesn't do the right thing with epoch rollover yet
    gitdir = feedpath / 'git' / f'{latest_epoch}.git'
    commits = lei.get_latest_commits_in_epoch(gitdir)
    if commits:
        logger.debug('Found %d new commits for list %s', len(commits), listname)
        count, last_commit = process_commits(listname=listname, commits=commits,
                                             gitdir=gitdir, ctx=ctx, max_count=max_mail)
        lei.save_epoch_info(list_dir=feedpath, epochs=latest_epochs)
        return count
    else:
        logger.debug('No new commits to process for LEI list %s', listname)
        return 0


def process_lore_list(ctx: click.Context, listname: str,
                      details: Dict[str, Any], max_mail: int) -> int:
    ls = ctx.obj['lore']
    if ls is None:
        data_dir = ctx.obj['data_dir']
        ls = LoreService(data_dir)
        ctx.obj['lore'] = ls
    latest_epochs = ls.get_epochs(details['feed'])
    count = 0

    data_dir = ctx.obj['data_dir']

    list_dir = data_dir / f'{listname}'
    if not list_dir.exists():
        ls.init_list(list_name=listname, list_dir=list_dir, pi_url=details['feed'])
        ls.store_epochs_info(list_dir=list_dir, epochs=latest_epochs)
        logger.info('Initialized: %s.', listname)
        return 0

    current_epochs: List[Tuple[int, str, str]] = list()
    try:
        current_epochs = ls.load_epochs_info(list_dir=list_dir)
    except StateError:
        pass

    if current_epochs == latest_epochs:
        logger.debug('No updates for lore list: %s', listname)
        return 0

    # Pull the highest epoch we have
    logger.debug('Running git pull on list: %s', listname)
    highest_epoch, gitdir, commits = ls.pull_highest_epoch(list_dir=list_dir)
    if commits:
        logger.debug('Found %d new commits for list %s', len(commits), listname)
        count, last_commit = process_commits(listname=listname, commits=commits,
                                             gitdir=gitdir, ctx=ctx, max_count=max_mail)
    else:
        last_commit = ''
        logger.debug('No new commits to process for list %s', listname)

    local = set(e[0] for e in current_epochs)
    remote = set(e[0] for e in latest_epochs)

    new_epochs = remote - local
    if new_epochs:
        # In theory, we could have more than one new epoch, for example if
        # someone hasn't run korgalore in a long time. This is almost certainly
        # not something anyone would want, because it would involve pulling a lot of data
        # that would take ages. So for now, we just pick the highest new epoch, which
        # will be correct in vast majority of cases.
        next_epoch = max(new_epochs)
        repo_url = f"{details['feed'].rstrip('/')}/git/{next_epoch}.git"
        tgt_dir = list_dir / 'git' / f'{next_epoch}.git'
        logger.debug('Cloning new epoch %d for list %s', next_epoch, listname)
        ls.clone_epoch(repo_url=repo_url, tgt_dir=tgt_dir, shallow=False)
        commits = ls.get_all_commits_in_epoch(tgt_dir)
        # attempt to respect max_mail across epoch boundaries
        remaining_mail = max_mail - count if max_mail > 0 else 0
        if remaining_mail <= 0:
            # Not clear what to do in this case, so we're just going to do max_mail for
            # the new epoch as well
            remaining_mail = max_mail
        new_count, last_commit = process_commits(listname=listname, commits=commits,
                                                 gitdir=tgt_dir, ctx=ctx, max_count=remaining_mail)
        count += new_count

    ls.store_epochs_info(list_dir=list_dir, epochs=latest_epochs)

    return count


@click.group()
@click.version_option(version=__version__)
@click_log.simple_verbosity_option(logger)
@click.option('--cfgfile', '-c', help='Path to configuration file.')
@click.option('-l', '--logfile', default=None, type=click.Path(), help='Path to log file.')
@click.pass_context
def main(ctx: click.Context, cfgfile: str, logfile: Optional[click.Path]) -> None:
    ctx.ensure_object(dict)

    # Load configuration file
    if not cfgfile:
        cfgdir = get_xdg_config_dir()
        cfgpath = cfgdir / 'korgalore.toml'
    else:
        cfgpath = Path(cfgfile)

    if logfile:
        file_handler = logging.FileHandler(str(logfile))
        file_handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # Only load config if we're not in edit-config mode
    if ctx.invoked_subcommand != 'edit-config':
        config = load_config(cfgpath)
        ctx.obj['config'] = config

    # Ensure XDG data directory exists
    data_dir = get_xdg_data_dir()
    ctx.obj['data_dir'] = data_dir

    logger.debug('Data directory: %s', data_dir)

    # We lazy-load these services as needed
    ctx.obj['targets'] = dict()

    ctx.obj['lore'] = None
    ctx.obj['lei'] = None


@main.command()
@click.pass_context
def auth(ctx: click.Context) -> None:
    """Authenticate with Gmail."""
    config = ctx.obj.get('config', {})
    targets = config.get('targets', {})
    if not targets:
        logger.critical('No targets defined in configuration.')
        raise click.Abort()
    for identifier, details in targets.items():
        if details.get('type') != 'gmail':
            continue
        get_gmail_service(identifier=identifier,
                          credentials_file=details.get('credentials', ''),
                          token_file=details.get('token', None))
    logger.info('Authentication complete.')


@main.command()
@click.pass_context
def edit_config(ctx: click.Context) -> None:
    """Open the configuration file in the default editor."""
    # Get config file path
    cfgfile = ctx.parent.params.get('cfgfile') if ctx.parent else None
    if not cfgfile:
        cfgdir = get_xdg_config_dir()
        cfgpath = cfgdir / 'korgalore.toml'
    else:
        cfgpath = Path(cfgfile)

    # Create config file with example if it doesn't exist
    if not cfgpath.exists():
        logger.info('Configuration file does not exist. Creating example configuration at: %s', cfgpath)
        example_config = """### Targets ###

[targets.personal]
type = 'gmail'
credentials = '~/.config/korgalore/credentials.json'
# token = '~/.config/korgalore/token.json'

### Sources ###

# [sources.lkml]
# feed = 'https://lore.kernel.org/lkml'
# target = 'personal'
# labels = ['INBOX', 'UNREAD']
"""
        cfgpath.parent.mkdir(parents=True, exist_ok=True)
        cfgpath.write_text(example_config)

    # Open in editor
    logger.info('Editing configuration file: %s', cfgpath)
    click.edit(filename=str(cfgpath))
    logger.debug('Configuration file closed.')


@main.command()
@click.pass_context
@click.argument('target', type=str, nargs=1)
@click.option('--ids', '-i', is_flag=True, help='include id values')
def labels(ctx: click.Context, target: str, ids: bool = False) -> None:
    """List all available labels."""
    gs = get_target(ctx, ctx.params['target'])

    try:
        logger.debug('Fetching labels from Gmail')
        labels_list = gs.list_labels()

        if not labels_list:
            logger.info("No labels found.")
            return

        logger.debug('Found %d labels', len(labels_list))
        logger.info('Available labels:')
        for label in labels_list:
            if ids:
                logger.info(f"  - {label['name']} (ID: {label['id']})")
            else:
                logger.info(f"  - {label['name']}")

    except Exception as e:
        logger.critical('Failed to fetch labels: %s', str(e))
        raise click.Abort()


@main.command()
@click.pass_context
@click.option('--max-mail', '-m', default=0, help='maximum number of messages to pull (0 for all)')
@click.argument('listname', type=str, nargs=1, default=None)
def pull(ctx: click.Context, max_mail: int, listname: Optional[str]) -> None:
    cfg = ctx.obj.get('config', {})

    sources = cfg.get('sources', {})
    if listname:
        if listname not in sources:
            logger.critical('List "%s" not found in configuration.', listname)
            raise click.Abort()
        sources = {listname: sources[listname]}

    changes: List[Tuple[str, int]] = list()
    for listname, details in sources.items():
        logger.debug('Processing list: %s', listname)
        if details.get('feed', '').startswith('https:'):
            try:
                count = process_lore_list(ctx=ctx, listname=listname, details=details, max_mail=max_mail)
            except Exception as e:
                logger.critical('Failed to process lore list "%s": %s', listname, str(e))
                continue
            if count > 0:
                changes.append((listname, count))
        elif details.get('feed', '').startswith('lei:'):
            try:
                count = process_lei_list(ctx=ctx, listname=listname, details=details, max_mail=max_mail)
            except Exception as e:
                logger.critical('Failed to process LEI list "%s": %s', listname, str(e))
                continue
            if count > 0:
                changes.append((listname, count))
        else:
            logger.warning('Unknown feed type for list %s: %s', listname, details.get('feed'))
            continue
    if changes:
        logger.info('Pull complete with updates:')
        for listname, count in changes:
            logger.info('  %s: %d', listname, count)
    else:
        logger.info('Pull complete with no updates.')


if __name__ == '__main__':
    main()



import json
import logging

from email.message import EmailMessage
from email.parser import BytesParser
from email.policy import EmailPolicy
from email import charset
from pathlib import Path
from korgalore import PublicInboxError, GitError, StateError

from typing import Any, Dict, List, Optional, Tuple, Union

from datetime import datetime

charset.add_charset('utf-8', None)
logger = logging.getLogger('korgalore')

class PIService:
    GITCMD: str = "git"
    emlpolicy: EmailPolicy = EmailPolicy(utf8=True, cte_type='8bit', max_line_length=None,
                                         message_factory=EmailMessage)

    def __init__(self) -> None:
        pass

    def run_git_command(self, topdir: Optional[str], args: List[str]) -> Tuple[int, bytes]:
        """Run a git command in the specified topdir and return (returncode, output)."""
        import subprocess

        cmd = [self.GITCMD]
        if topdir:
            cmd += ['-C', topdir]
        cmd += args
        logger.debug('Running git command: %s', ' '.join(cmd))

        try:
            result = subprocess.run(cmd, capture_output=True)
        except FileNotFoundError:
            raise GitError(f"Git command '{self.GITCMD}' not found. Is it installed?")
        return result.returncode, result.stdout.strip()

    def find_epochs(self, topdir: Path) -> List[int]:
        epochs_dir = topdir / 'git'
        # List this directory for existing epochs
        existing_epochs: List[int] = list()
        for item in epochs_dir.iterdir():
            if item.is_dir() and item.name.endswith('.git'):
                epoch_str = item.name.replace('.git', '')
                try:
                    epoch_num = int(epoch_str)
                    existing_epochs.append(epoch_num)
                except ValueError:
                    logger.debug(f"Invalid epoch directory: {item.name}")
        if not existing_epochs:
            raise PublicInboxError(f"No existing epochs found in {epochs_dir}.")
        return sorted(existing_epochs)

    def get_all_commits_in_epoch(self, gitdir: Path) -> List[str]:
        gitargs = ['rev-list', '--reverse', 'master']
        retcode, output = self.run_git_command(str(gitdir), gitargs)
        if retcode != 0:
            raise GitError(f"Git rev-list failed: {output.decode()}")
        if len(output):
            commits = output.decode().splitlines()
        else:
            commits = []
        return commits

    def recover_after_rebase(self, tgt_dir: Path) -> str:
        # Load korgalore.info to find last processed commit
        info = self.load_korgalore_info(tgt_dir)
        # Get the commit's date and parse it into datetime
        # The string is ISO with tzinfo: "2025-11-04 20:47:21 +0000"
        commit_date_str = info.get('commit_date')
        if not commit_date_str:
            raise StateError(f"No commit_date found in korgalore.info in {tgt_dir}.")
        commit_date = datetime.strptime(commit_date_str, '%Y-%m-%d %H:%M:%S %z')
        logger.debug(f"Last processed commit date: {commit_date.isoformat()}")
        # Try to find the new hash of this commit in the log by matching the subject and
        # message-id.
        gitargs = ['rev-list', '--reverse', '--since-as-filter', commit_date_str, 'master']
        retcode, output = self.run_git_command(str(tgt_dir), gitargs)
        if retcode != 0:
            # Not sure what happened here, just give up and return the latest commit
            logger.warning("Could not run rev-list to recover after rebase, returning latest commit.")
            latest_commit = self.get_top_commit(tgt_dir)
            return latest_commit

        possible_commits = output.decode().splitlines()
        if not possible_commits:
            # Just record the latest info, then
            self.update_korgalore_info(gitdir=tgt_dir)
            latest_commit = self.get_top_commit(tgt_dir)
            return latest_commit

        last_commit = ''
        first_commit = possible_commits[0]
        for commit in possible_commits:
            raw_message = self.get_message_at_commit(tgt_dir, commit)
            msg = self.parse_message(raw_message)
            subject = msg.get('Subject', '(no subject)')
            msgid = msg.get('Message-ID', '(no message-id)')
            if subject == info.get('subject') and msgid == info.get('msgid'):
                logger.debug(f"Found matching commit: {commit}")
                last_commit = commit
                break
        if not last_commit:
            logger.error("Could not find exact commit after rebase.")
            logger.error("Returning first possible commit after date: %s", first_commit)
            last_commit = first_commit
            raw_message = self.get_message_at_commit(tgt_dir, last_commit)
            msg = self.parse_message(raw_message)
        else:
            logger.debug("Recovered exact matching commit after rebase: %s", last_commit)

        self.update_korgalore_info(gitdir=tgt_dir, latest_commit=last_commit, message=msg)
        return last_commit

    def get_latest_commits_in_epoch(self, gitdir: Path,
                                    since_commit: Optional[str] = None) -> List[str]:
        # How many new commits since our latest_commit
        if not since_commit:
            try:
                info = self.load_korgalore_info(gitdir)
            except StateError:
                raise StateError(f"korgalore.info not found in {gitdir}. Run init_list() first.")
            since_commit = info.get('last')
        # is this still a valid commit?
        gitargs = ['cat-file', '-e', f'{since_commit}^']
        retcode, output = self.run_git_command(str(gitdir), gitargs)
        if retcode != 0:
            # The commit is not valid anymore, so try to find the latest commit by other
            # means.
            logger.debug(f"Since commit {since_commit} not found, trying to recover after rebase.")
            since_commit = self.recover_after_rebase(gitdir)
        gitargs = ['rev-list', '--reverse', '--ancestry-path', f'{since_commit}..master']
        retcode, output = self.run_git_command(str(gitdir), gitargs)
        if retcode != 0:
            raise GitError(f"Git rev-list failed: {output.decode()}")
        if len(output):
            new_commits = output.decode().splitlines()
        else:
            new_commits = []
        return new_commits

    def get_message_at_commit(self, pi_dir: Path, commitish: str) -> bytes:
        gitargs = ['show', f'{commitish}:m']
        retcode, output = self.run_git_command(str(pi_dir), gitargs)
        if retcode == 128:
            raise StateError(f"Commit {commitish} does not have a message file.")
        if retcode != 0:
            raise GitError(f"Git show failed: {output.decode()}")
        return output

    def parse_message(self, raw_message: bytes) -> EmailMessage:
        """Parse a raw email message into an EmailMessage object."""
        msg: EmailMessage = BytesParser(_class=EmailMessage,
                                        policy=self.emlpolicy).parsebytes(raw_message)  # type: ignore
        return msg

    def get_top_commit(self, gitdir: Path) -> str:
        gitargs = ['rev-list', '-n', '1', 'master']
        retcode, output = self.run_git_command(str(gitdir), gitargs)
        if retcode != 0:
            raise GitError(f"Git rev-list failed: {output.decode()}")
        top_commit = output.decode()
        return top_commit

    def update_korgalore_info(self, gitdir: Path,
                              latest_commit: Optional[str] = None,
                              message: Optional[Union[bytes, EmailMessage]] = None) -> None:
        if not latest_commit:
            latest_commit = self.get_top_commit(gitdir)

        # Get the commit date
        gitargs = ['show', '-s', '--format=%ci', latest_commit]
        retcode, output = self.run_git_command(str(gitdir), gitargs)
        if retcode != 0:
            raise GitError(f"Git show failed: {output.decode()}")
        commit_date = output.decode()
        # TODO: latest_commit may not have a "m" file in it if it's a deletion
        korgalore_file = Path(gitdir) / 'korgalore.info'
        if not message:
            message = self.get_message_at_commit(gitdir, latest_commit)

        if isinstance(message, bytes):
            msg = self.parse_message(message)
        else:
            msg = message
        subject = msg.get('Subject', '(no subject)')
        msgid = msg.get('Message-ID', '(no message-id)')
        with open(korgalore_file, 'w') as gf:
            json.dump({
                'last': latest_commit,
                'subject': subject,
                'msgid': msgid,
                'commit_date': commit_date,
            }, gf, indent=2)

    def load_korgalore_info(self, gitdir: Path) -> Dict[str, Any]:
        korgalore_file = Path(gitdir) / 'korgalore.info'
        if not korgalore_file.exists():
            raise StateError(
                f"korgalore.info not found in {gitdir}. Run init_list() first."
            )

        with open(korgalore_file, 'r') as gf:
            info = json.load(gf)  # type: Dict[str, Any]

        return info




import json
import logging
from pathlib import Path
from typing import List, Tuple

from korgalore.pi_service import PIService
from korgalore import GitError, PublicInboxError, StateError

logger = logging.getLogger('korgalore')

class LeiService(PIService):
    """Service for interacting with lore.kernel.org public-inbox archives."""
    LEICMD: str = "lei"

    def __init__(self) -> None:
        # do a parent init
        super().__init__()
        self.known_searches: List[str] = list()
        self._load_known_searches()

    def run_lei_command(self, args: List[str]) -> Tuple[int, bytes]:
        import subprocess

        cmd = [self.LEICMD]
        cmd += args

        try:
            result = subprocess.run(cmd, capture_output=True)
        except FileNotFoundError:
            raise PublicInboxError(f"LEI command '{self.LEICMD}' not found. Is it installed?")
        return result.returncode, result.stdout.strip()

    def get_latest_epoch_info(self, list_dir: Path) -> List[Tuple[int, str]]:
        epochs = self.find_epochs(list_dir)
        epoch_info: List[Tuple[int, str]] = list()
        for epoch in epochs:
            epoch_dir = list_dir / 'git' / f'{epoch}.git'
            gitargs = ['show-ref']
            retcode, output = self.run_git_command(str(epoch_dir), gitargs)
            if retcode != 0:
                raise GitError(f"Git show-ref failed: {output.decode()}")
            # It's just one ref in lei repos
            refdata = output.decode()
            logger.debug('Epoch %d refdata: %s', epoch, refdata)
            epoch_info.append((epoch, refdata))
        return epoch_info

    def load_known_epoch_info(self, list_dir: Path) -> List[Tuple[int, str]]:
        epochs_file = list_dir / 'epochs.json'
        if not epochs_file.exists():
            raise StateError(f"Epochs file {epochs_file} does not exist.")
        with open(epochs_file, 'r') as ef:
            epochs_data = json.load(ef)
        epochs: List[Tuple[int, str]] = list()
        for entry in epochs_data:
            epochs.append((entry['epoch'], entry['refdata']))
        return epochs

    def save_epoch_info(self, list_dir: Path, epochs: List[Tuple[int, str]]) -> None:
        epochs_file = list_dir / 'epochs.json'
        epochs_info = list()
        for enum, refdata in epochs:
            epochs_info.append({
                'epoch': enum,
                'refdata': refdata
            })
        with open(epochs_file, 'w') as ef:
            json.dump(epochs_info, ef, indent=2)

    def _load_known_searches(self) -> None:
        args = ['ls-search', '-l', '-f', 'json']
        retcode, output = self.run_lei_command(args)
        if retcode != 0:
            raise PublicInboxError(f"LEI list searches failed: {output.decode()}")
        json_output = output.decode()
        ls_data = json.loads(json_output)
        # Only return the names of v2 searches
        for entry in ls_data:
            output = entry.get('output', '')
            if output.startswith('v2:'):
                self.known_searches.append(output[3:])

    def up_search(self, lei_name: str) -> None:
        leiargs = ['up', lei_name]
        retcode, output = self.run_lei_command(leiargs)
        if retcode != 0:
            raise PublicInboxError(f"LEI update failed: {output.decode()}")


import requests
from typing import List, Dict, Tuple, Any
from gzip import GzipFile
from pathlib import Path
from email import charset
import io
import json

import logging

from korgalore import __version__, StateError, RemoteError
from korgalore.pi_service import PIService

charset.add_charset('utf-8', None)

logger = logging.getLogger('korgalore')


class LoreService(PIService):
    """Service for interacting with lore.kernel.org public-inbox archives."""

    def __init__(self, datadir: Path) -> None:
        # do a parent init
        super().__init__()
        self.session = requests.Session()
        self.session.headers.update({
            'User-Agent': f'korgalore/{__version__}'
        })
        self.datadir = datadir

    def get_manifest(self, pi_url: str) -> Dict[str, Any]:
        try:
            response = self.session.get(f"{pi_url.rstrip('/')}/manifest.js.gz")
            response.raise_for_status()
        except Exception as e:
            raise RemoteError(
                f"Failed to fetch manifest from {pi_url}: {e}"
            ) from e
        # ungzip and parse the manifest
        manifest: Dict[str, Any] = dict()
        with GzipFile(fileobj=io.BytesIO(response.content)) as f:
            mf = json.load(f)
            for key, vals in mf.items():
                manifest[key] = vals

        return manifest

    def clone_epoch(self, repo_url: str, tgt_dir: Path, shallow: bool = True) -> None:
        # does tgt_dir exist?
        if Path(tgt_dir).exists():
            logger.debug(f"Target directory {tgt_dir} already exists, skipping clone.")
            return

        gitargs = ['clone', '--mirror']
        if shallow:
            gitargs += ['--shallow-since=1.week.ago']
        gitargs += [repo_url, str(tgt_dir)]

        retcode, output = self.run_git_command(None, gitargs)
        if retcode != 0:
            raise RemoteError(f"Git clone failed: {output.decode()}")

    def get_epochs(self, pi_url: str) -> List[Tuple[int, str, str]]:
        manifest = self.get_manifest(pi_url)
        # The keys are epoch paths, so we extract epoch numbers and paths
        epochs: List[Tuple[int, str, str]] = []
        # The key ends in #.git, so grab the final path component and remove .git
        for epoch_path in manifest.keys():
            epoch_str = epoch_path.split('/')[-1].replace('.git', '')
            try:
                epoch_num = int(epoch_str)
                fpr = str(manifest[epoch_path]['fingerprint'])
                epochs.append((epoch_num, epoch_path, fpr))
            except ValueError:
                logger.warning(f"Invalid epoch string: {epoch_str} in {pi_url}")
        # Sort epochs by their numeric value
        epochs.sort(key=lambda x: x[0])
        return epochs

    def store_epochs_info(self, list_dir: Path, epochs: List[Tuple[int, str, str]]) -> None:
        epochs_file = list_dir / 'epochs.json'
        epochs_info = []
        for enum, epath, fpr in epochs:
            epochs_info.append({
                'epoch': enum,
                'path': epath,
                'fpr': fpr
            })
        with open(epochs_file, 'w') as ef:
            json.dump(epochs_info, ef, indent=2)

    def load_epochs_info(self, list_dir: Path) -> List[Tuple[int, str, str]]:
        epochs_file = list_dir / 'epochs.json'
        if not epochs_file.exists():
            raise StateError(f"Epochs file {epochs_file} does not exist.")
        with open(epochs_file, 'r') as ef:
            epochs_data = json.load(ef)
        epochs: List[Tuple[int, str, str]] = []
        for entry in epochs_data:
            epochs.append((entry['epoch'], entry['path'], entry['fpr']))
        return epochs

    def init_list(self, list_name: str, list_dir: Path, pi_url: str) -> None:
        if not list_dir.exists():
            list_dir.mkdir(parents=True, exist_ok=True)
        epochs = self.get_epochs(pi_url)
        enum, epath, _ = epochs[-1]
        tgt_dir = list_dir / 'git' / f'{enum}.git'
        repo_url = f"{pi_url.rstrip('/')}/git/{enum}.git"
        self.clone_epoch(repo_url=repo_url, tgt_dir=tgt_dir)
        self.update_korgalore_info(gitdir=tgt_dir)

    def pull_highest_epoch(self, list_dir: Path) -> Tuple[int, Path, List[str]]:
        # What is our highest epoch?
        existing_epochs = self.find_epochs(list_dir)
        highest_epoch = max(existing_epochs)
        logger.debug(f"Highest epoch found: {highest_epoch}")
        epochs_dir = list_dir / 'git'
        tgt_dir = epochs_dir / f'{highest_epoch}.git'
        # Pull the latest changes
        gitargs = ['fetch', 'origin', '--shallow-since=1.week.ago', '--update-shallow']
        retcode, output = self.run_git_command(str(tgt_dir), gitargs)
        if retcode != 0:
            raise RemoteError(f"Git remote update failed: {output.decode()}")
        new_commits = self.get_latest_commits_in_epoch(tgt_dir)
        return highest_epoch, tgt_dir, new_commits



"""Korgalore - A command-line tool to put public-inbox sources directly into Gmail."""

__version__ = "0.1.1"
__author__ = "Konstantin Ryabitsev"
__email__ = "konstantin@linuxfoundation.org"


# Custom exceptions
class KorgaloreError(Exception):
    """Base exception for all Korgalore errors."""
    pass


class ConfigurationError(KorgaloreError):
    """Raised when there is an error in configuration."""
    pass


class GitError(KorgaloreError):
    """Raised when there is an error with Git operations."""
    pass


class RemoteError(KorgaloreError):
    """Raised when there is an error communicating with remote services."""
    pass

class PublicInboxError(KorgaloreError):
    """Raised when something is wrong with Public-Inbox."""
    pass

class StateError(KorgaloreError):
    """Raised when there is an error with the internal state."""
    pass



import os
import logging
from typing import Optional, List, Dict, Any

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow # type: ignore
from googleapiclient.discovery import build # type: ignore
from googleapiclient.errors import HttpError # type: ignore

from korgalore import ConfigurationError, RemoteError

logger = logging.getLogger('korgalore')

# If modifying these scopes, delete the file token.json.
# We need scopes for reading and inserting new emails, but not
# modifying existing ones.
SCOPES = [
    'https://www.googleapis.com/auth/gmail.labels',
    'https://www.googleapis.com/auth/gmail.insert',
    ]


class GmailService:
    def __init__(self, identifier: str, credentials_file: str, token_file: str) -> None:
        self.identifier = identifier
        self.creds: Optional[Credentials] = None
        self.service: Optional[Any] = None
        self._load_credentials(credentials_file, token_file)
        self._label_map: Optional[Dict[str, str]] = None

    def _load_credentials(self, credentials_file: str, token_file: str) -> None:
        # Expand vars and tildes on file paths
        credentials_file = os.path.expandvars(os.path.expanduser(credentials_file))
        token_file = os.path.expandvars(os.path.expanduser(token_file))
        # The file token.json stores the user's access and refresh tokens
        if os.path.exists(token_file):
            self.creds = Credentials.from_authorized_user_file(token_file, SCOPES) # type: ignore

        # If there are no (valid) credentials available, let the user log in
        if not self.creds or not self.creds.valid:
            if self.creds and self.creds.expired and self.creds.refresh_token:
                self.creds.refresh(Request())  # type: ignore
            elif os.path.exists(credentials_file):
                logger.critical('Log in to Gmail account for %s', self.identifier)

                flow = InstalledAppFlow.from_client_secrets_file(
                    credentials_file, SCOPES)
                self.creds = flow.run_local_server(port=0)
            else:
                raise ConfigurationError(
                    f"{credentials_file} not found. Please download it from Google Cloud Console."
                )

            # Save the credentials for the next run
            with open(token_file, 'w') as token:
                token.write(self.creds.to_json())

        self.service = build('gmail', 'v1', credentials=self.creds)

    def list_labels(self) -> List[Dict[str, str]]:
        """List all labels in the user's mailbox.

        Returns:
            List of label objects
        """
        try:
            results = self.service.users().labels().list(userId='me').execute()  # type: ignore
            labels = results.get('labels', [])
            return labels  # type: ignore

        except HttpError as error:
            raise RemoteError(f'An error occurred: {error}')

    def translate_labels(self, labels: List[str]) -> List[str]:
        # Translate label names to their corresponding IDs
        if self._label_map is None:
            # Get all labels from Gmail
            self._label_map = {label['name']: label['id'] for label in self.list_labels()}
        translated: List[str] = []
        for label in labels:
            label_id = self._label_map.get(label, None)
            if label_id is None:
                raise ConfigurationError(f"Label '{label}' not found in Gmail '{self.identifier}'.")
            translated.append(label_id)
        return translated

    def import_message(self, raw_message: bytes, labels: List[str]) -> Any:
        try:
            import base64

            encoded_message = base64.urlsafe_b64encode(raw_message).decode()
            message_body: Dict[str, Any] = {'raw': encoded_message}

            if labels:
                label_ids = self.translate_labels(labels)
                message_body['labelIds'] = label_ids

            # Upload the message
            result = self.service.users().messages().import_(  # type: ignore
                userId='me',
                body=message_body
            ).execute()

            return result

        except HttpError as error:
            raise RemoteError(f'An error occurred: {error}')
